@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 
 """
-Copyright (C) 2008-2012 Wolfgang Rohdewald <wolfgang@rohdewald.de>
+Copyright (C) 2008-2014 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
-kajongg is free software you can redistribute it and/or modify
+Kajongg is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 2 of the License, or
 (at your option) any later version.
@@ -18,22 +18,24 @@ along with this program if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
-from PyQt4.QtCore import Qt, QPointF, QPoint, QRectF, QMimeData, QSize, QVariant
-from PyQt4.QtGui import QGraphicsRectItem, QGraphicsItem, QSizePolicy, QFrame, QFont
-from PyQt4.QtGui import QGraphicsView, QGraphicsEllipseItem, QGraphicsScene, QLabel
-from PyQt4.QtGui import QColor, QPainter, QDrag, QPixmap, QStyleOptionGraphicsItem, QPen, QBrush
-from PyQt4.QtGui import QFontMetrics, QTransform
-from PyQt4.QtGui import QMenu, QCursor
-from PyQt4.QtSvg import QGraphicsSvgItem
+from qt import usingQt5, Qt, QPointF, QPoint, QRectF, QMimeData, QSize, QVariant
+from qt import QGraphicsRectItem, QGraphicsItem, QSizePolicy, QFrame, QFont
+from qt import QGraphicsView, QGraphicsEllipseItem, QLabel
+from qt import QColor, QPainter, QDrag, QPixmap, QStyleOptionGraphicsItem, QPen, QBrush
+from qt import QFontMetrics, QTransform
+from qt import QMenu, QCursor
+from qt import QGraphicsSvgItem
 from tileset import Tileset, TileException
-from tile import chiNext, Tile, elements
+from tile import Tile, elements
 from uitile import UITile, UIMeld
+from guiutil import Painter
 from meld import Meld
 from animation import Animation, Animated, animate
 from message import Message
 
-from util import logDebug, logException, m18n, m18nc, kprint, stack, uniqueList
-from common import WINDS, LIGHTSOURCES, Internal, ZValues, Debug, Preferences, isAlive
+from util import kprint, stack, uniqueList
+from log import logDebug, logException, m18n, m18nc
+from common import WINDS, LIGHTSOURCES, Internal, Debug, isAlive
 
 ROUNDWINDCOLOR = QColor(235, 235, 173)
 
@@ -49,7 +51,7 @@ def rotateCenter(item, angle):
 
 class PlayerWind(QGraphicsEllipseItem):
     """a round wind tile"""
-    def __init__(self, name, tileset, roundsFinished=0, parent = None):
+    def __init__(self, name, tileset, roundsFinished=0, parent=None):
         """generate new wind tile"""
         if not len(WINDPIXMAPS):
             WINDPIXMAPS[('E', False)] = None # avoid recursion
@@ -68,7 +70,7 @@ class PlayerWind(QGraphicsEllipseItem):
     @staticmethod
     def genWINDPIXMAPS():
         """prepare wind tiles"""
-        tileset = Tileset(Preferences.windTilesetName)
+        tileset = Tileset(Internal.Preferences.windTilesetName)
         for wind in WINDS:
             for prevailing in False, True:
                 pwind = PlayerWind(wind, tileset, prevailing)
@@ -80,10 +82,9 @@ class PlayerWind(QGraphicsEllipseItem):
                 pwind.paint(painter, QStyleOptionGraphicsItem())
                 for child in pwind.childItems():
                     if isinstance(child, QGraphicsSvgItem):
-                        painter.save()
-                        painter.translate(child.mapToParent(0.0, 0.0))
-                        child.paint(painter, QStyleOptionGraphicsItem())
-                        painter.restore()
+                        with Painter(painter):
+                            painter.translate(child.mapToParent(0.0, 0.0))
+                            child.paint(painter, QStyleOptionGraphicsItem())
                 WINDPIXMAPS[(wind, prevailing)] = pMap
 
     def __sizeFace(self):
@@ -127,7 +128,7 @@ class WindLabel(QLabel):
             self.__wind = wind
             self._refresh()
 
-    def __init__(self, wind = None, roundsFinished = 0, parent=None):
+    def __init__(self, wind=None, roundsFinished=0, parent=None):
         QLabel.__init__(self, parent)
         self.__wind = None
         if wind is None:
@@ -160,7 +161,7 @@ class Board(QGraphicsRectItem):
     arrows = [Qt.Key_Left, Qt.Key_Down, Qt.Key_Up, Qt.Key_Right]
     def __init__(self, width, height, tileset, boardRotation=0):
         QGraphicsRectItem.__init__(self)
-        self.tiles = []
+        self.uiTiles = []
         self.isHandBoard = False
         self._focusTile = None
         self.__prevPos = 0
@@ -168,34 +169,41 @@ class Board(QGraphicsRectItem):
         self.tileDragEnabled = False
         self.setRotation(boardRotation)
         self._lightSource = 'NW'
-        self.xWidth = 0
-        self.xHeight = 0
-        self.yWidth = 0
-        self.yHeight = 0
+        self.__xWidth = 0
+        self.__xHeight = 0
+        self.__yWidth = 0
+        self.__yHeight = 0
         self.__fixedWidth = width
         self.__fixedHeight = height
         self._tileset = None
         self._showShadows = None
         self.tileset = tileset
         self.level = 0
+        self.showShadows = Internal.Preferences.showShadows
 
-    # pylint: disable=no-self-use
-    def name(self):
+    @property
+    def name(self): # pylint: disable=no-self-use
         """default board name, used for debugging messages"""
         return 'board'
 
+    def setVisible(self, value):
+        """also update focusRect if it belongs to this board"""
+        if self.scene() and isAlive(self):
+            QGraphicsRectItem.setVisible(self, value)
+            self.scene().focusRect.refresh()
+
     def hide(self):
-        """remove all tile references so they can be garbage collected"""
-        for tile in self.tiles:
-            tile.hide()
-        self.tiles = []
+        """remove all uiTile references so they can be garbage collected"""
+        for uiTile in self.uiTiles:
+            uiTile.hide()
+        self.uiTiles = []
         self._focusTile = None
         if isAlive(self):
-            QGraphicsRectItem.hide(self)
+            self.setVisible(False)
 
     def autoSelectTile(self):
-        """call this when kajongg should automatically focus
-        on an appropriate tile"""
+        """call this when Kajongg should automatically focus
+        on an appropriate uiTile"""
         focusCandidates = self._focusableTiles()
         if focusCandidates:
             firstCandidate = focusCandidates[0]
@@ -211,32 +219,27 @@ class Board(QGraphicsRectItem):
 
     @property
     def focusTile(self):
-        """the tile of this board with focus. This is per Board!"""
+        """the uiTile of this board with focus. This is per Board!"""
         if self._focusTile is None:
             self.autoSelectTile()
         return self._focusTile
 
     @focusTile.setter
-    def focusTile(self, tile):
-        """the tile of this board with focus. This is per Board!"""
-        if tile is self._focusTile:
+    def focusTile(self, uiTile):
+        """the uiTile of this board with focus. This is per Board!"""
+        if uiTile is self._focusTile:
             return
-        if tile:
-            assert tile.tile != 'Xy', tile
-            if not isinstance(tile.board, DiscardBoard):
-                assert tile.focusable, tile
-            self.__prevPos = tile.sortKey()
-        self._focusTile = tile
+        if uiTile:
+            assert uiTile.tile.isKnown, uiTile
+            if not isinstance(uiTile.board, DiscardBoard):
+                assert uiTile.focusable, uiTile
+            self.__prevPos = uiTile.sortKey()
+        self._focusTile = uiTile
         if self._focusTile and self._focusTile.tile in Debug.focusable:
-            logDebug('%s: new focus tile %s from %s' % (
-                self.name(), self._focusTile.tile if self._focusTile else 'None', stack('')[-1]))
-        if (self.isHandBoard and self.player
-            and not self.player.game.isScoringGame()
-            and Internal.field.clientDialog):
-            Internal.field.clientDialog.focusTileChanged()
+            logDebug('%s: new focus uiTile %s from %s' % (
+                self.name, self._focusTile.tile if self._focusTile else 'None', stack('')[-1]))
         if self.hasFocus:
             self.scene().focusBoard = self
-        assert tile is self._focusTile
 
     def setEnabled(self, enabled):
         """enable/disable this board"""
@@ -249,7 +252,7 @@ class Board(QGraphicsRectItem):
         the next list element.
         respect board orientation: Right Arrow should always move right
         relative to the screen, not relative to the board"""
-        return sorted([x for x in self.tiles if x.focusable], key=lambda x: x.sortKey(sortDir))
+        return sorted([x for x in self.uiTiles if x.focusable], key=lambda x: x.sortKey(sortDir))
 
     @property
     def hasFocus(self):
@@ -263,10 +266,12 @@ class Board(QGraphicsRectItem):
         """sets focus on this board"""
         if isAlive(self):
             scene = self.scene()
-            if scene.focusBoard == self or value:
-                if self.focusTile:
-                    assert self.focusTile.board == self, '%s not in self %s' % (self.focusTile, self)
-                scene.focusBoard = self if value else None
+            if isAlive(scene):
+                if scene.focusBoard == self or value:
+                    if self.focusTile:
+                        assert self.focusTile.board == self, '%s not in self %s' % (self.focusTile, self)
+                if value:
+                    scene.focusBoard = self
 
     @staticmethod
     def mapChar2Arrow(event):
@@ -297,17 +302,22 @@ class Board(QGraphicsRectItem):
             # Original Call.
             oldPos = self.focusTile.xoffset, self.focusTile.yoffset
             tiles = list(x for x in tiles if (x.xoffset, x.yoffset) != oldPos or x == self.focusTile)
-            assert tiles, [str(x) for x in self.tiles]
+            assert tiles, [str(x) for x in self.uiTiles]
             tiles.append(tiles[0])
             self.focusTile = tiles[tiles.index(self.focusTile)+1]
 
-    def uiMeldWithTile(self, uiTile):
+    def mapMouseTile(self, uiTile): # pylint: disable=no-self-use
+        """map the pressed tile to the wanted tile. For melds, this would
+        be the first tile no matter which one is pressed"""
+        return uiTile
+
+    def uiMeldWithTile(self, uiTile): # pylint: disable=no-self-use
         """returns the UI Meld with uiTile. A Board does not know about melds,
         so default is to return a Meld with only uiTile"""
         return Meld(uiTile)
 
-    def meldVariants(self, uiTile, lowerHalf): # pylint: disable=unused-argument
-        """all possible melds that could be meant by dragging/dropping tile"""
+    def meldVariants(self, uiTile, lowerHalf): # pylint: disable=no-self-use,unused-argument
+        """all possible melds that could be meant by dragging/dropping uiTile"""
         return [Meld(uiTile)]
 
     def chooseVariant(self, uiTile, lowerHalf=False):
@@ -320,16 +330,19 @@ class Board(QGraphicsRectItem):
             for idx, variant in enumerate(variants):
                 action = menu.addAction(variant.typeName())
                 action.setData(QVariant(idx))
-            if Internal.field.centralView.dragObject:
+            if Internal.scene.mainWindow.centralView.dragObject:
                 menuPoint = QCursor.pos()
             else:
                 menuPoint = uiTile.board.tileFaceRect().bottomRight()
-                view = Internal.field.centralView
+                view = Internal.scene.mainWindow.centralView
                 menuPoint = view.mapToGlobal(view.mapFromScene(uiTile.mapToScene(menuPoint)))
             action = menu.exec_(menuPoint)
             if not action:
                 return None
-            idx = action.data().toInt()[0]
+            if usingQt5:
+                idx = action.data() # autoconvert
+            else:
+                idx = action.data().toInt()[0]
         return variants[idx]
 
     def dragEnterEvent(self, dummyEvent):
@@ -345,23 +358,23 @@ class Board(QGraphicsRectItem):
         self.setPen(QPen(Qt.NoPen))
 
     def tileAt(self, xoffset, yoffset, level=0):
-        """if there is a tile at this place, return it"""
-        for tile in self.tiles:
-            if (tile.xoffset, tile.yoffset, tile.level) == (xoffset, yoffset, level):
-                return tile
+        """if there is a uiTile at this place, return it"""
+        for uiTile in self.uiTiles:
+            if (uiTile.xoffset, uiTile.yoffset, uiTile.level) == (xoffset, yoffset, level):
+                return uiTile
 
     def tilesByElement(self, element):
-        """returns all child items holding a tile for element"""
-        return list(x for x in self.tiles if x.tile == element)
+        """returns all child items holding a uiTile for element"""
+        return list(x for x in self.uiTiles if x.tile is element)
 
     def rotatedLightSource(self):
-        """the light source we need for the original tile before it is rotated"""
+        """the light source we need for the original uiTile before it is rotated"""
         lightSourceIndex = LIGHTSOURCES.index(self.lightSource)
         lightSourceIndex = (lightSourceIndex+self.sceneRotation() // 90)%4
         return LIGHTSOURCES[lightSourceIndex]
 
     def tileFacePos(self):
-        """the face pos of a tile relative to the tile origin"""
+        """the face pos of a uiTile relative to its origin"""
         if not self.showShadows:
             return QPointF()
         lightSource = self.rotatedLightSource()
@@ -370,7 +383,7 @@ class Board(QGraphicsRectItem):
         return QPointF(xoffset, yoffset)
 
     def tileFaceRect(self):
-        """the face rect of a tile relative to the tile origin"""
+        """the face rect of a uiTile relative its origin"""
         return QRectF(self.tileFacePos(), self.tileset.faceSize)
 
     def sceneRotation(self):
@@ -385,14 +398,14 @@ class Board(QGraphicsRectItem):
     def setPos(self, xWidth=0, xHeight=0, yWidth=0, yHeight=0):
         """sets the position in the parent item expressing the position in tile face units.
         The X position is xWidth*facewidth + xHeight*faceheight, analog for Y"""
-        self.xWidth = xWidth
-        self.xHeight = xHeight
-        self.yWidth = yWidth
-        self.yHeight = yHeight
+        self.__xWidth = xWidth
+        self.__xHeight = xHeight
+        self.__yWidth = yWidth
+        self.__yHeight = yHeight
         self.setGeometry()
 
     def setRect(self, width, height):
-        """gives the board a fixed size in tile coordinates"""
+        """gives the board a fixed size in uiTile coordinates"""
         self.__fixedWidth = width
         self.__fixedHeight = height
         self.computeRect()
@@ -432,8 +445,8 @@ class Board(QGraphicsRectItem):
             offsets = (-self.tileset.shadowHeight() * 2, 0)
         else:
             offsets = self.tileset.shadowOffsets(self._lightSource, self.sceneRotation())
-        newX = self.xWidth*width+self.xHeight*height + offsets[0]
-        newY = self.yWidth*width+self.yHeight*height + offsets[1]
+        newX = self.__xWidth*width+self.__xHeight*height + offsets[0]
+        newY = self.__yWidth*width+self.__yHeight*height + offsets[1]
         QGraphicsRectItem.setPos(self, newX, newY)
 
     @property
@@ -445,8 +458,8 @@ class Board(QGraphicsRectItem):
     def showShadows(self, value):
         """set active lightSource"""
         if self._showShadows != value:
-            for tile in self.tiles:
-                tile.setClippingFlags()
+            for uiTile in self.uiTiles:
+                uiTile.setClippingFlags()
             self._reload(self.tileset, showShadows=value)
 
     @property
@@ -491,9 +504,9 @@ class Board(QGraphicsRectItem):
                     child.tileset = tileset
                     child.lightSource = lightSource
                     child.showShadows = showShadows
-            for tile in self.tiles:
-                self.placeTile(tile)
-                tile.update()
+            for uiTile in self.uiTiles:
+                self.placeTile(uiTile)
+                uiTile.update()
             self.computeRect()
             if self.hasFocus:
                 self.scene().focusBoard = self
@@ -503,9 +516,9 @@ class Board(QGraphicsRectItem):
         return 1
 
     def shiftZ(self, level):
-        """used for 3D: compute the needed shift for the tile.
+        """used for 3D: compute the needed shift for the uiTile.
         level is the vertical position. 0 is the face position on
-        ground level, -1 is the imprint a tile makes on the
+        ground level, -1 is the imprint a uiTile makes on the
         surface it stands on"""
         if not self.showShadows:
             return QPointF()
@@ -526,66 +539,66 @@ class Board(QGraphicsRectItem):
         return QPointF(shiftX, shiftY)
 
     def tileSize(self):
-        """the current tile size"""
+        """the current uiTile size"""
         return self._tileset.tileSize
 
     def faceSize(self):
         """the current face size"""
         return self._tileset.faceSize
 
-    def __tilePlace(self, tile):
-        """compute all properties for tile in this board: pos, scale, rotation
+    def __tilePlace(self, uiTile):
+        """compute all properties for uiTile in this board: pos, scale, rotation
         and return them in a dict"""
-        assert isinstance(tile, UITile)
-        if not tile.scene():
-            self.scene().addItem(tile)
+        assert isinstance(uiTile, UITile)
+        if not uiTile.scene():
+            self.scene().addItem(uiTile)
         width = self.tileset.faceSize.width()
         height = self.tileset.faceSize.height()
-        shiftZ = self.shiftZ(tile.level)
-        boardPos = QPointF(tile.xoffset*width, tile.yoffset*height) + shiftZ
+        shiftZ = self.shiftZ(uiTile.level)
+        boardPos = QPointF(uiTile.xoffset*width, uiTile.yoffset*height) + shiftZ
         scenePos = self.mapToScene(boardPos)
-        tile.setDrawingOrder()
+        uiTile.setDrawingOrder()
         return {'pos': scenePos, 'rotation': self.sceneRotation(), 'scale': self.scale()}
 
-    def placeTile(self, tile):
-        """places the tile in the scene. With direct=False, animate"""
-        assert isinstance(tile, UITile)
-        for pName, newValue in self.__tilePlace(tile).items():
-            animation = tile.queuedAnimation(pName)
+    def placeTile(self, uiTile):
+        """places the uiTile in the scene. With direct=False, animate"""
+        assert isinstance(uiTile, UITile)
+        for pName, newValue in self.__tilePlace(uiTile).items():
+            animation = uiTile.queuedAnimation(pName)
             if animation:
                 curValue = animation.unpackValue(animation.endValue())
                 if curValue != newValue:
                     # change a queued animation
                     animation.setEndValue(newValue)
             else:
-                animation = tile.activeAnimation.get(pName, None)
+                animation = uiTile.activeAnimation.get(pName, None)
                 if isAlive(animation):
                     curValue = animation.unpackValue(animation.endValue())
                 else:
-                    curValue = tile.getValue(pName)
+                    curValue = uiTile.getValue(pName)
                 if pName != 'scale' or abs(curValue - newValue) > 0.00001:
                     if curValue != newValue:
-                        Animation(tile, pName, newValue)
+                        Animation(uiTile, pName, newValue)
 
     def addUITile(self, uiTile):
         """add uiTile to this board"""
-        self.tiles.append(uiTile)
+        self.uiTiles.append(uiTile)
 
     def removeUITile(self, uiTile):
         """remove uiTile from this board"""
-        self.tiles.remove(uiTile)
+        self.uiTiles.remove(uiTile)
         if self.currentFocusTile == uiTile:
             self.focusTile = None
 
 class CourtBoard(Board):
     """A Board that is displayed within the wall"""
     def __init__(self, width, height):
-        Board.__init__(self, width, height, Internal.field.tileset)
+        Board.__init__(self, width, height, Internal.scene.tileset)
         self.setAcceptDrops(True)
 
     def maximize(self):
         """make it as big as possible within the wall"""
-        cWall = Internal.field.game.wall
+        cWall = Internal.scene.game.wall
         newSceneX = cWall[3].sceneBoundingRect().right()
         newSceneY = cWall[2].sceneBoundingRect().bottom()
         tileset = self.tileset
@@ -607,8 +620,8 @@ class CourtBoard(Board):
         yScaleFactor = yAvail / yNeeded
         QGraphicsRectItem.setPos(self, newSceneX, newSceneY)
         Board.setScale(self, min(xScaleFactor, yScaleFactor))
-        for tile in self.tiles:
-            tile.board.placeTile(tile)
+        for uiTile in self.uiTiles:
+            uiTile.board.placeTile(uiTile)
 
 class SelectorBoard(CourtBoard):
     """a board containing all possible tiles for selection"""
@@ -624,50 +637,50 @@ class SelectorBoard(CourtBoard):
 
     def load(self, game):
         """load the tiles according to game.ruleset"""
-        for tile in self.tiles:
-            tile.setBoard(None)
-        self.tiles = []
+        for uiTile in self.uiTiles:
+            uiTile.setBoard(None)
+        self.uiTiles = []
         self.allSelectorTiles = list(UITile(x) for x in elements.all(game.ruleset))
         self.refill()
 
     def refill(self):
         """move all tiles back into the selector"""
         with Animated(False):
-            for tileItem in self.allSelectorTiles:
-                tileItem.tile = Tile(tileItem.tile.lower())
-                self.__placeAvailable(tileItem)
-                tileItem.dark = False
-                tileItem.focusable = True
+            for uiTile in self.allSelectorTiles:
+                uiTile.tile = uiTile.tile.exposed
+                self.__placeAvailable(uiTile)
+                uiTile.dark = False
+                uiTile.focusable = True
             self.focusTile = self.tilesByElement(Tile('c1'))[0]
 
-    # pylint: disable=no-self-use
-    def name(self):
+    @property
+    def name(self): # pylint: disable=no-self-use
         """for debugging messages"""
         return 'selector'
 
     def dragMoveEvent(self, event):
         """allow dropping only from handboards, not from self"""
-        tile = event.mimeData().tile
-        event.setAccepted(tile.board != self)
+        uiTile = event.mimeData().uiTile
+        event.setAccepted(uiTile.board != self)
 
     def dropEvent(self, event):
-        """drop a tile into the selector"""
-        tile = event.mimeData().tile
-        self.dropTile(tile)
+        """drop a uiTile into the selector"""
+        uiTile = event.mimeData().uiTile
+        self.dropTile(uiTile)
         event.accept()
 
-    def dropTile(self, tile, lowerHalf=False): # pylint: disable=unused-argument
-        """drop tile or meld into selector board"""
-        senderHand = tile.board
+    def dropTile(self, uiTile, lowerHalf=False): # pylint: disable=unused-argument
+        """drop uiTile into selector board"""
+        senderHand = uiTile.board
         if senderHand == self:
             return
-        meld = tile.board.uiMeldWithTile(tile)
-        self.lastReceived = tile
+        meld = uiTile.board.uiMeldWithTile(uiTile)
+        self.lastReceived = uiTile
         for myTile in meld:
             self.__placeAvailable(myTile)
             myTile.focusable = True
         senderHand.deselect(meld)
-        (senderHand if senderHand.tiles else self).hasFocus = True
+        (senderHand if senderHand.uiTiles else self).hasFocus = True
         self._noPen()
         animate()
 
@@ -676,66 +689,64 @@ class SelectorBoard(CourtBoard):
         assert isinstance(uiTile, UITile), uiTile
         result = UIMeld(uiTile)
         for tile in meld[1:]:
-            baseTiles = list(x for x in self.tilesByElement(tile.lower()) if x not in result)
+            baseTiles = list(x for x in self.tilesByElement(tile.exposed) if x not in result)
             result.append(baseTiles[0])
         return result
 
     def deselect(self, meld):
         """we are going to lose those tiles or melds"""
 
-    def __placeAvailable(self, tile):
-        """place the tile in the selector at its place"""
+    def __placeAvailable(self, uiTile):
+        """place the uiTile in the selector at its place"""
         # define coordinates and order for tiles:
-        offsets = {'d': (3, 6, 'bgr'), 'f': (4, 5, 'eswn'), 'y': (4, 0, 'eswn'),
-            'w': (3, 0, 'eswn'), 'b': (1, 0, '123456789'), 's': (2, 0, '123456789'),
-            'c': (0, 0, '123456789')}
-        row, baseColumn, order = offsets[tile.tile[0].lower()]
-        column = baseColumn + order.index(tile.tile[1])
-        tile.dark = False
-        tile.setBoard(self, column, row)
+        offsets = {Tile.dragon: (3, 6, Tile.dragons), Tile.flower: (4, 5, Tile.winds), Tile.season: (4, 0, Tile.winds),
+            Tile.wind: (3, 0, Tile.winds), Tile.bamboo: (1, 0, Tile.numbers), Tile.stone: (2, 0, Tile.numbers),
+            Tile.character: (0, 0, Tile.numbers)}
+        row, baseColumn, order = offsets[uiTile.tile.lowerGroup]
+        column = baseColumn + order.index(uiTile.tile.value)
+        uiTile.dark = False
+        uiTile.setBoard(self, column, row)
 
     def meldVariants(self, uiTile, lowerHalf):
         """returns a list of possible variants based on meld. Those are logical melds."""
         # pylint: disable=too-many-locals
         assert isinstance(uiTile, UITile)
         wantedTile = uiTile.tile
-        for selectorTile in self.tiles:
-            selectorTile.tile = selectorTile.tile.lower()
-        lowerName = wantedTile.lower()
-        upperName = wantedTile.upper()
+        for selectorTile in self.uiTiles:
+            selectorTile.tile = selectorTile.tile.exposed
+        lowerName = wantedTile.exposed
+        upperName = wantedTile.concealed
         if lowerHalf:
             scName = upperName
         else:
             scName = lowerName
-        variants = [scName]
+        result = [scName.single]
         baseTiles = len(self.tilesByElement(lowerName))
         if baseTiles >= 2:
-            variants.append(scName * 2)
+            result.append(scName.pair)
         if baseTiles >= 3:
-            variants.append(scName * 3)
+            result.append(scName.pung)
         if baseTiles == 4:
             if lowerHalf:
-                variants.append(lowerName + upperName * 2 + lowerName)
+                result.append(lowerName.kong.declared)
             else:
-                variants.append(lowerName * 4)
-                variants.append(lowerName * 3 + upperName)
-        if not wantedTile.isHonor() and wantedTile[1] < '8':
-            chow2 = chiNext(wantedTile, 1)
-            chow3 = chiNext(wantedTile, 2)
-            chow2 = self.tilesByElement(chow2.lower())
-            chow3 = self.tilesByElement(chow3.lower())
-            if chow2 and chow3:
-                baseChar = scName[0]
-                baseValue = ord(scName[1])
-                varStr = '%s%s%s%s%s' % (scName, baseChar, chr(baseValue+1), baseChar, chr(baseValue+2))
-                variants.append(varStr)
-        return [Meld(x) for x in variants]
+                result.append(lowerName.kong)
+                result.append(lowerName.kong.exposedClaimed)
+        # TODO: old bug: move 3 identical hidden chows to east, then move
+        # one of them to south exposed
+        if not wantedTile.isHonor and wantedTile.value < 8:
+            chow2 = scName.nextForChow
+            chow3 = chow2.nextForChow
+            if self.tilesByElement(chow2.exposed) and self.tilesByElement(chow3.exposed):
+                result.append(scName.chow)
+        # result now holds a list of melds
+        return result
 
 class MimeData(QMimeData):
     """we also want to pass a reference to the moved meld"""
     def __init__(self, uiTile):
         QMimeData.__init__(self)
-        self.tile = uiTile
+        self.uiTile = uiTile
         self.setText(str(uiTile))
 
 class FittingView(QGraphicsView):
@@ -761,7 +772,7 @@ class FittingView(QGraphicsView):
         """we do not want scrolling for the scene view.
         Instead scrolling down changes perspective like in kmahjongg"""
         if event.orientation() == Qt.Vertical and event.delta() < 0:
-            Internal.field.changeAngle()
+            Internal.scene.changeAngle()
         # otherwise do not call ignore() because we do want
         # to consume this
 
@@ -773,9 +784,6 @@ class FittingView(QGraphicsView):
         if parent:
             grandpa = parent.parentWidget()
             if grandpa and grandpa.objectName() == 'MainWindow':
-                if grandpa.ignoreResizing:
-                    grandpa.ignoreResizing -= 1
-                    return
                 grandpa.applySettings()
                 # resize background:
                 grandpa.backgroundName = grandpa.backgroundName
@@ -783,15 +791,15 @@ class FittingView(QGraphicsView):
             self.fitInView(self.scene().itemsBoundingRect(), Qt.KeepAspectRatio)
         self.setFocus()
 
-    def __matchingTile(self, position, item):
-        """is position in the clickableRect of this tile?"""
-        if not isinstance(item, UITile):
+    def __matchingTile(self, position, uiTile):
+        """is position in the clickableRect of this uiTile?"""
+        if not isinstance(uiTile, UITile):
             return False
-        itemPos = item.mapFromScene(self.mapToScene(position))
-        return item.board.tileFaceRect().contains(itemPos)
+        itemPos = uiTile.mapFromScene(self.mapToScene(position))
+        return uiTile.board.tileFaceRect().contains(itemPos)
 
     def tileAt(self, position):
-        """find out which tile is clickable at this position. Always
+        """find out which uiTile is clickable at this position. Always
         returns a list. If there are several tiles above each other,
         return all of them, highest first"""
         allTiles = [x for x in self.items(position) if isinstance(x, UITile)]
@@ -810,19 +818,17 @@ class FittingView(QGraphicsView):
         tiles = self.tileAt(event.pos())
         if tiles:
             if event.modifiers() & Qt.ShiftModifier:
-                for tile in tiles:
-                    kprint('%s: board.level:%s' % (str(tile), tile.board.level))
-            tile = tiles[0]
-            board = tile.board
-            isRemote = board.isHandBoard and board.player and not board.player.game.isScoringGame()
-            if board.isHandBoard and not isRemote and not tile.isBonus():
-                tile = tile.board.uiMeldWithTile(tile)[0]
-            if tile.focusable:
-                board.focusTile = tile
+                for uiTile in tiles:
+                    kprint('%s: board.level:%s' % (str(uiTile), uiTile.board.level))
+            board = tiles[0].board
+            uiTile = board.mapMouseTile(tiles[0])
+            if uiTile.focusable:
+                board.focusTile = uiTile
                 board.hasFocus = True
-                if isRemote:
-                    Internal.field.clientDialog.buttons[0].setFocus()
-                self.tilePressed = tile
+                if hasattr(Internal.scene, 'clientDialog'):
+                    if Internal.scene.clientDialog:
+                        Internal.scene.clientDialog.buttons[0].setFocus()
+                self.tilePressed = uiTile
             else:
                 event.ignore()
         else:
@@ -835,7 +841,7 @@ class FittingView(QGraphicsView):
         return QGraphicsView.mouseReleaseEvent(self, event)
 
     def mouseMoveEvent(self, event):
-        """selects the correct tile"""
+        """selects the correct uiTile"""
         tilePressed = self.tilePressed
         self.tilePressed = None
         if tilePressed:
@@ -847,15 +853,15 @@ class FittingView(QGraphicsView):
                 return
         return QGraphicsView.mouseMoveEvent(self, event)
 
-    def drag(self, tile):
+    def drag(self, uiTile):
         """returns a drag object"""
         drag = QDrag(self)
-        mimeData = MimeData(tile)
+        mimeData = MimeData(uiTile)
         drag.setMimeData(mimeData)
-        tRect = tile.boundingRect()
+        tRect = uiTile.boundingRect()
         tRect = self.viewportTransform().mapRect(tRect)
-        pmapSize = QSize(tRect.width() * tile.scale, tRect.height() * tile.scale)
-        pMap = tile.pixmapFromSvg(pmapSize)
+        pmapSize = QSize(tRect.width() * uiTile.scale, tRect.height() * uiTile.scale)
+        pMap = uiTile.pixmapFromSvg(pmapSize)
         drag.setPixmap(pMap)
         drag.setHotSpot(QPoint(pMap.width()/2, pMap.height()/2))
         return drag
@@ -880,18 +886,15 @@ class YellowText(QGraphicsRectItem):
         self.height = metrics.lineSpacing() * 1.1
         self.setRect(0, 0, self.width, self.height)
         self.resetTransform()
+        self.setPos(self.side.center())
         rotation = self.side.rotation()
         rotateCenter(self, -rotation)
+        xOffset = -self.rect().width()/2
+        yOffset = -self.rect().height()/2
         if rotation % 180 == 0:
-            yOffset = self.rect().height()
-            if rotation == 0:
-                yOffset = 2 * -yOffset
-            if rotation == 180:
-                self.translate(self.rect().width()/2, yOffset)
-            else:
-                self.translate(-self.rect().width()/2, yOffset)
+            self.moveBy(xOffset, yOffset * 4)
         else:
-            self.translate(-self.rect().width()/2, -self.rect().height()/2)
+            self.moveBy(xOffset, yOffset)
     def paint(self, painter, dummyOption, dummyWidget):
         """override predefined paint"""
         painter.setFont(self.font)
@@ -900,18 +903,19 @@ class YellowText(QGraphicsRectItem):
 
 class DiscardBoard(CourtBoard):
     """A special board for discarded tiles"""
+    # pylint: disable=too-many-public-methods
     def __init__(self):
         CourtBoard.__init__(self, 11, 9)
         self.__places = None
         self.lastDiscarded = None
 
-    @staticmethod
-    def name(): # pylint: disable=arguments-differ
+    @property
+    def name(self): # pylint: disable=no-self-use
         """to be used in debug output"""
         return "discardBoard"
 
     def hide(self):
-        """remove all tile references so they can be garbage collected"""
+        """remove all uiTile references so they can be garbage collected"""
         self.lastDiscarded = None
         CourtBoard.hide(self)
 
@@ -920,121 +924,23 @@ class DiscardBoard(CourtBoard):
         self.__places = [(x, y) for x in range(self.width) for y in range(self.height)]
         randomGenerator.shuffle(self.__places)
 
-    def discardTile(self, tile):
-        """add tile to a random position"""
-        assert isinstance(tile, UITile)
-        tile.setBoard(self, *self.__places.pop(0))
-        tile.dark = False
-        tile.focusable = False
-        self.focusTile = tile
+    def discardTile(self, uiTile):
+        """add uiTile to a random position"""
+        assert isinstance(uiTile, UITile)
+        uiTile.setBoard(self, *self.__places.pop(0))
+        uiTile.dark = False
+        uiTile.focusable = False
+        self.focusTile = uiTile
         self.hasFocus = True
-        self.lastDiscarded = tile
+        self.lastDiscarded = uiTile
 
     def dropEvent(self, event):
-        """drop a tile into the discard board"""
+        """drop a uiTile into the discard board"""
         # TODO: now that tiles are top level scene items, maybe drag them
         # directly. Draggings melds: QGraphicsItemGroup?
-        tile = event.mimeData().tile
-        assert isinstance(tile, UITile), tile
-        tile.setPos(event.scenePos() - tile.boundingRect().center())
-        Internal.field.clientDialog.selectButton(Message.Discard)
+        uiTile = event.mimeData().uiTile
+        assert isinstance(uiTile, UITile), uiTile
+        uiTile.setPos(event.scenePos() - uiTile.boundingRect().center())
+        Internal.scene.clientDialog.selectButton(Message.Discard)
         event.accept()
         self._noPen()
-
-class MJScene(QGraphicsScene):
-    """our scene with a potential Qt bug fix"""
-    def __init__(self):
-        QGraphicsScene.__init__(self)
-        self.__disableFocusRect = False
-        self._focusBoard = None
-        self.focusRect = QGraphicsRectItem()
-        pen = QPen(QColor(Qt.blue))
-        pen.setWidth(6)
-        self.focusRect.setPen(pen)
-        self.addItem(self.focusRect)
-        self.focusRect.setZValue(ZValues.marker)
-        self.focusRect.hide()
-
-    def focusInEvent(self, event):
-        """work around a qt bug. See https://bugreports.qt-project.org/browse/QTBUG-32890
-        This can be reproduced as follows:
-           ./kajongg.py --game=whatever --autoplay=SomeRuleset
-               such that the human player is the first one to discard a tile.
-           wait until the main screen has been built
-           click with the mouse into the middle of that window
-           press left arrow key
-           this will violate the assertion in UITile.keyPressEvent """
-        prev = self.focusItem()
-        QGraphicsScene.focusInEvent(self, event)
-        if prev and bool(prev.flags() & QGraphicsItem.ItemIsFocusable) and prev != self.focusItem():
-            self.setFocusItem(prev)
-
-    def __focusRectVisible(self):
-        """should we show it?"""
-        game = Internal.field.game
-        board = self._focusBoard
-        return bool(not self.__disableFocusRect
-                and board
-                and board.isVisible()
-                and board.isEnabled()
-                and board.hasFocus
-                and board.focusTile
-                and game
-                and not game.autoPlay)
-
-    @property
-    def disableFocusRect(self):
-        """suppress focusrect"""
-        return self.__disableFocusRect
-
-    @disableFocusRect.setter
-    def disableFocusRect(self, value):
-        """always place or hide, even if value does not change"""
-        self.__disableFocusRect = value
-        if value:
-            self.focusRect.hide()
-        else:
-            self.placeFocusRect()
-
-    @property
-    def focusBoard(self):
-        """get / set the board that has its focusRect shown"""
-        return self._focusBoard
-
-    @focusBoard.setter
-    def focusBoard(self, board):
-        """get / set the board that has its focusRect shown"""
-        self._focusBoard = board
-        focusTile = board.focusTile if board else None
-        if focusTile:
-            focusTile.setFocus()
-            self.placeFocusRect()
-        self.focusRect.setVisible(self.__focusRectVisible())
-
-    def placeFocusRect(self):
-        """show a blue rect around tile"""
-        board = self._focusBoard
-        if board and board.isVisible() and board.focusTile:
-            rect = board.tileFaceRect()
-            rect.setWidth(rect.width()*board.focusRectWidth())
-            self.focusRect.setRect(rect)
-            self.focusRect.setPos(board.focusTile.pos)
-            self.focusRect.setRotation(board.sceneRotation())
-            self.focusRect.setScale(board.scale())
-            self.focusRect.show()
-        else:
-            self.focusRect.hide()
-
-    def graphicsTileItems(self):
-        """returns all UITile in the scene"""
-        return (x for x in self.items() if isinstance(x, UITile))
-
-    def nonTiles(self):
-        """returns all other items in the scene"""
-        return (x for x in self.items() if not isinstance(x, UITile))
-
-    def removeTiles(self):
-        """remove all tiles from scene"""
-        for item in self.graphicsTileItems():
-            self.removeItem(item)
-        self.focusRect.hide()
